import { Constant } from '../../decorators/constant';
import * as _ from 'lodash';

@Constant('objectUtils', true)

/**
 * @class ObjectUtils
 *
 * Utilities for iterable objects (key/value objects and arrays).
 * Comparable to what underscore calls 'collections'
 */
export class ObjectUtils {
  /**
   * Remove an item from an array and return the supplied array
   *
   * @todo move to array utils
   *
   * @param  {Array} arr   Array to remove the item from
   * @param  {any}   item  Item to remove from the array
   * @return {Array}       The array supplied, now with the given item removed
   */
  remove(arr: any[], item: any): any[] {
    let index = 0;

    while (index !== -1) {
      index = arr.indexOf(item, index);
      if (index !== -1) {
        arr.splice(index, 1);
      }
    }

    return arr;
  }

  /**
   * Remove all key/value pairs from an object or array
   *
   * @param  {Object|any[]} obj
   * Object to clear
   *
   * @return {Object}           Originally supplied object
   */
  clear(obj: any): any {

    // Arrays
    if (_.isArray(obj)) {
      obj.length = 0;
    }

    // Objects
    for (const key in obj) {
      delete obj[key];
    }

    return obj;
  }

  /**
   * Replace all keys/values of one object/array with another
   * but keep the reference to the original
   *
   * @param  {Object|any[]} obj1
   * Source object to clear and extend with the second
   *
   * @param  {Object|any[]} obj2
   * New object to extend all key/values to the source
   *
   * @return {Object|any[]}
   *  Modified source object supplied
   */
  replace(obj1: any, obj2: any): any {
    this.clear(obj1);
    return _.extend(obj1, obj2);
  }

  /**
   * Recursively removes any undefined values from an object.
   */
  prune(object: any) {
    _.forEach(object, (value: any, key: string) => {
       if (_.isObject(value)) {
        value = object[key] = this.prune(value);
       }

      if (!(_.isBoolean(value) || _.isNumber(value)) && _.isEmpty(value)) {
        delete object[key];
      }
    })

    return object
  }

  /**
   *  Flatten an object, e.g.
   *
   *   from { user: { name: 'steve' } }
   *   to   { 'user.name': 'steve' }
   *
   * Credit: http://stackoverflow.com/a/19101235
   *  */
  flatten(object: any, flattenArrays = true) {
    const result = {};

    const recurse = (cur: any, prop: string): void => {
      if (Object(cur) !== cur) {
        result[prop] = cur;
      } else if (Array.isArray(cur)) {
        if (flattenArrays) {
          let i = 0;
          let l = cur.length;

          while(i < l) {
            recurse(cur[i], prop + '[' + i + ']');
            i++;
          }

          if (l === 0) {
            result[prop] = [];
          } else {
            result[prop] = cur;
          }
        }
      } else {
        let isEmpty = true;

        for (const p in cur) {
          isEmpty = false;
          recurse(cur[p], prop ? prop + '.' + p : p);
          if (isEmpty && prop) {
            result[prop] = {};
          }
        }
      }
    }

    recurse(object, '');
    return result;
  }
}
