import '@host';

import { Provider }  from '../decorators/provider';
import { inject }    from '../decorators/inject';
import { Component } from '../services/component';
import { Lazy }      from '../services/lazy';
import * as _        from 'lodash';
// import 'oclazyload';

/**
 * @interface IState
 * @extends angular.ui.IState
 * @description
 * Extends angular.ui.IState to support a comopnent property
 * that takes a component name and turns it into a template
 */
export interface IState extends angular.ui.IState {
  component?: string;

  /**
   * If there is a component for this state, should it be
   * lazy loaed?
   *
   * @type {Boolean}
   * @default false
   */
  lazy?: Boolean;

  resolve?: any;

  // TODO: just call 'view'?
  attachToView?: string;
  wrapTemplate?: boolean;

  /**
   * Source module to find component
   * @example
   *   module: '@ssjs/shopstyle-components'
   */
  module?: string;

  /**
   * Path to component
   */
  path?: string;

  hideNavBar?: boolean;
}

export interface IRouterDefaults {

  /**
   * Default route to redirect to if no others match
   * @type {string}
   */
  defaultRoute: string;

  /**
   * List of states
   * @type {IState[]}
   * ```
   */
  routes: IState[];

  /**
   * String to prepend to all component state names. E.g.
   * with the default 'app' a component by the name 'home'
   * wil lhave the state name 'app.home'
   * @todo way to configure these as a provider
   * @default 'app'
   * @type {String}
   */
  componentStatePrefix: string;


  /**
   * Name of main ui-view. This is what is used
   * for the view name all components nest under.
   * E.g. for 'menuContent' component routes get const ructed as
   *
   * @type {String}
   * @default 'menuContent';
   */
  mainViewName: string;
}


@Provider('router')

/**
 * @class Routes
 * @description
 * Configures and const ructs routes/states via angular.ui.router
 * @example
 * ```ts
 *
 * export class Routes extends Router {
 * 	constructor() {
 * 		super();
 * 		this.applyRoutes();
 * 	}
 *
 * 	defaultRoute: '/';
 *
 * 	routes: [
 *   	{
 *   		component: 'home',
 *   		url: '/'
 * 		},
 * 		{
 * 			component: 'test'
 * 		}
 * 	]
 * }
 * ```
 */
export class Router implements IRouterDefaults {

  /*
   *injections
   */
  @inject
  $stateProvider: angular.ui.IStateProvider;

  @inject
  $urlRouterProvider: angular.ui.IUrlRouterProvider;

  /*
   * Property defaults
   */
  componentStatePrefix = 'app';
  mainViewName = 'menuContent';
  defaultRoute = '/';
  routes: IState[] = [];

  static noop() {
    // do nothing
  }

  constructor() {
    // Initialize
    Lazy.noop();
  }

  /**
   * Set router defaults via IRouterDefaults interface
   *
   * @param {IRouterDefaults} options   Router defaults to set
   * @param {boolean}         defaults  Only set if not already set
   */
  set(options: IRouterDefaults, defaults = false): void {
    if (defaults) {
      _.defaults(options, this);
    } else {
      _.extend(options, this);
    }
  }


  // TODO: class Router extends SingletonProvider
  $get(): Router {
    return this;
  }


  /**
   * Takes a state object and const ructs and applies the route
   * for it via $stateProvider.state()
   *
   * @param  {IState}    state  State object to apply
   * @return {ui.IState}        Converted state object (to angular.ui.IState format)
   */
  applyRoute(state: IState): angular.ui.IState {
    const uiState = this.configureState(state);
    this.$stateProvider.state(state.name, uiState);
    return uiState;
  }


  /**
   * Converts an IState to a angular.ui.IState
   *
   * @description
   * Note that any property can be overwritten, so you can
   * supply custom urls, views, component name, etc and the
   * 'component' property will fill in any missing fields
   *
   * @param  {IState} state  Default state object to conert
   * @return {ui.IState}     Converted state object
   *
   * @example
   * Converts
   * ```ts
   * { component: 'home' }
   * ```
   * to
   * ```ts
   * {
   * 	name: 'app.home',
   * 	url: '/home',
   * 	views: {
   * 		menuContent: {
   * 			template: ```
   * 				<ion-view title="Home">
   * 					<ion-content>
   * 						<home></home>
   * 					</ion-content>
   * 				</ion-view>
   * 			```
   * 		}
   * 	}
   * }
   * ```
   */
  configureState(state: IState): angular.ui.IState {
    const { mainViewName, componentStatePrefix } = this;
    const { component } = state;
    const dashName = _.kebabCase(component);
    const camelName = _.camelCase(component);

    // Configure base state object, e.g.
    // {
    //  component: 'home'
    // }
    // to
    // {
    //  name: 'app.home'
    // }
    if (component) {

      // URL default
      if (_.isUndefined(state.url)) {
        state.url = `/${dashName}`;
      }

      // name default
      if (!state.name) {
        state.name = `${componentStatePrefix}.${camelName}`;
      }

      // wrap template default
      if (_.isUndefined(state.wrapTemplate)) {
        state.wrapTemplate = true;
      }

      if (!state.views) {
        const viewName = state.attachToView || mainViewName;
        state.views = {
          [viewName]: {
            component: component
          }
        };
      }

      // Enable lazy loading
      if (state.lazy !== false && !state.abstract) {
        state.resolve = state.resolve || {};

        // TODO: ngannotate for this
        state.resolve.$lazy = ['lazy', (lazy: Lazy) => {
          return lazy.load(component);
        }];
      }
    }

    // Configure views, e.g.
    // {
    //   component: 'main'
    // }
    // to
    // {
    //   template: '<ui-view title="Main">...</ui-view>'
    // }
    //
    for (const viewName in state.views) {
      const viewConfig = state.views[viewName];

      // If we see a component property turn it into
      // a template string instead via Component.viewTemplate
      if (viewConfig.component) {
        const componentInstance = new Component();

        if (state.wrapTemplate) {
          viewConfig.template = componentInstance.viewTemplate(viewConfig.component);
        } else {
          viewConfig.template = componentInstance.template(viewConfig.component);
        }
      }
    }

    return state;
  }

  /**
   * Apply routes configured via this.routes
   * @param {IState[]} routes List
   */
  applyRoutes(routes: IState[] = this.routes): void {
    if (this.routes) {
      for (const state of this.routes) {
        this.applyRoute(state);
      }
    }

    // Set the default route
    if (this.defaultRoute) {
      this.$urlRouterProvider.otherwise(this.defaultRoute);
    }
  }
}
